import { MiddlewareConsumer, Module } from '@nestjs/common';
import { ConfigModule, ConfigService } from '@nestjs/config';
import { APP_FILTER, APP_PIPE } from '@nestjs/core';
import { TypeOrmModule } from '@nestjs/typeorm';
import { I18nMiddleware, I18nModule } from 'nestjs-i18n';
import { LoggerModule } from 'nestjs-pino';
import { AllExceptionsFilter, buildI18nValidationExceptionFilter } from './core/filters';
import { buildConfigOptions, buildLoggerOptions, buildTypeormOptions } from './core/module-options';
import { buildI18nOptions } from './core/module-options/i18n-options';
import { buildValidationPipe } from './core/pipes';
import { migrations } from './db';
import { DocumentModule } from './document/document.module';
import { HealthModule } from './health/health.module';
@Module({
  controllers: [],
  imports: [
    ConfigModule.forRoot(buildConfigOptions()),
    TypeOrmModule.forRootAsync({
      imports: [],
      inject: [ConfigService],
      useFactory: (config: ConfigService) => buildTypeormOptions(config, migrations),
    }),
    LoggerModule.forRootAsync({
      useFactory: (config: ConfigService) => buildLoggerOptions(config),
      inject: [ConfigService],
    }),
    I18nModule.forRoot(buildI18nOptions()),
    HealthModule,

    // Application Modules
    DocumentModule,
  ],
  providers: [
    // Global Pipes
    {
      inject: [ConfigService],
      provide: APP_PIPE,
      useFactory: (config: ConfigService) => buildValidationPipe(config),
    },

    // Global Filters (order matters)
    { provide: APP_FILTER, useClass: AllExceptionsFilter },
    { provide: APP_FILTER, useValue: buildI18nValidationExceptionFilter() },
  ],
})
export class AppModule {
  configure(consumer: MiddlewareConsumer) {
    consumer.apply(I18nMiddleware).forRoutes('*');
  }
}
