import { ArgumentsHost, Catch, ExceptionFilter, HttpStatus, Logger } from '@nestjs/common';
import { HttpArgumentsHost } from '@nestjs/common/interfaces';
import { Response as ExpressResponse } from 'express';
import { I18nContext } from 'nestjs-i18n';
import { ErrorCategory } from '../enums';
import { IFieldError } from '../interfaces';
import { I18nContextWrapper, ResponseFactory } from '../utils';

const KNOWN_HTTP_EXCEPTION_CATEGORIES: { [_: number]: ErrorCategory } = {
  [HttpStatus.UNAUTHORIZED]: ErrorCategory.UNAUTHORIZED_ERROR,
  [HttpStatus.FORBIDDEN]: ErrorCategory.FORBIDDEN_ERROR,
  [HttpStatus.NOT_FOUND]: ErrorCategory.NOT_FOUND_ERROR,
  [HttpStatus.INTERNAL_SERVER_ERROR]: ErrorCategory.INTERNAL_SERVER_ERROR,
};

@Catch()
export class AllExceptionsFilter implements ExceptionFilter {
  private readonly logger = new Logger(AllExceptionsFilter.name);

  catch(exception: any, host: ArgumentsHost) {
    this.logger.error(exception);

    // Exception in HTTP context
    const httpCtx: HttpArgumentsHost = host.switchToHttp();
    const res = httpCtx.getResponse<ExpressResponse>();
    const i18n = new I18nContextWrapper(I18nContext.current());

    try {
      const status = this.extractStatusCode(exception);

      // Known categories (default errors, e.g. not found, unauthorized, ...)
      const knownCategory = KNOWN_HTTP_EXCEPTION_CATEGORIES[`${status}`];
      if (knownCategory) {
        const category = exception.response?.category || knownCategory;
        const message = i18n.ts(exception.response?.message, knownCategory);
        return res.status(status).json(ResponseFactory.error(category, message));
      }

      // Unknown Exception
      const category = exception.response?.category || ErrorCategory.UNHANDLED_ERROR;
      const message = this.extractMessage(exception);
      const errors = Array.isArray(exception.response?.errors) ? exception.response.errors : [];

      const resMessage = i18n.t(message);
      const resCategory = resMessage?.toString().includes('JSON') // Request body parsing error
        ? ErrorCategory.VALIDATION_ERROR
        : category;
      const resErrors = errors.map(
        (error: IFieldError) => ({ field: error.field, message: i18n.t(error.message) } as IFieldError),
      );

      return res.status(status).json(ResponseFactory.error(resCategory, resMessage, resErrors));
    } catch (err) {
      // Fallback
      this.logger.error(err);
      return res
        .status(HttpStatus.INTERNAL_SERVER_ERROR)
        .json(ResponseFactory.error(ErrorCategory.INTERNAL_SERVER_ERROR, i18n.t(ErrorCategory.INTERNAL_SERVER_ERROR)));
    }
  }

  private extractStatusCode(exception: any): number {
    return +(
      exception.getStatus?.() || // NestJS HttpException
      exception.response?.status ||
      exception.response?.statusCode ||
      exception.status ||
      HttpStatus.INTERNAL_SERVER_ERROR
    );
  }

  private extractMessage(exception: any): any {
    const messages = [
      exception.response?.data?.errorMessage,
      exception.response?.message, // NestJS HttpException
      exception.response?.error,
      exception.message,
    ];

    for (const message of messages) {
      if (message !== undefined) return message; // will accept falsy values except undefined (e.g. null, '', 0, ...)
    }

    return undefined;
  }
}
