import { buildTypeormOptions } from './typeorm-options';
import { ConfigService } from '@nestjs/config';
import { TypeOrmModuleOptions } from '@nestjs/typeorm';
import { Environment } from '../enums';
import { DeepMocked, createMock } from '@golevelup/ts-jest';

describe('buildTypeormOptions', () => {
  let configService: DeepMocked<ConfigService>;

  beforeEach(() => {
    configService = createMock<ConfigService>();

    jest.spyOn(configService, 'getOrThrow').mockImplementation((key) => {
      const envMap = {
        DB_HOST: 'localhost',
        DB_PORT: '5432',
        DB_USER: 'testuser',
        DB_PASS: 'testpass',
        DB_NAME: 'testdb',
        NODE_ENV: Environment.DEV,
        MIGRATIONS_RUN: 'true',
      };
      return envMap[key];
    });
  });

  it('should return correct TypeOrmModuleOptions with provided migrations', () => {
    const migrationsMock = ['src/migrations/*.ts'];

    const options: TypeOrmModuleOptions = buildTypeormOptions(configService, migrationsMock);

    expect(options).toEqual({
      type: 'postgres',
      host: 'localhost',
      port: 5432,
      username: 'testuser',
      password: 'testpass',
      database: 'testdb',
      logging: true,
      synchronize: false,
      migrationsRun: true,
      autoLoadEntities: true,
      migrations: migrationsMock,
    });
  });

  it('should set logging to false when NODE_ENV is not DEV', () => {
    jest
      .spyOn(configService, 'getOrThrow')
      .mockImplementation((key: string) => (key === 'NODE_ENV' ? Environment.PROD : undefined));

    const options: TypeOrmModuleOptions = buildTypeormOptions(configService);

    expect(options.logging).toBe(false);
  });

  it('should set migrationsRun to false when MIGRATIONS_RUN is not "true"', () => {
    jest
      .spyOn(configService, 'getOrThrow')
      .mockImplementation((key: string) => (key === 'MIGRATIONS_RUN' ? 'false' : undefined));

    const options: TypeOrmModuleOptions = buildTypeormOptions(configService);

    expect(options.migrationsRun).toBe(false);
  });
});
