import { BadRequestException, Injectable, Logger } from '@nestjs/common';
import { ConfigService } from '@nestjs/config';
import { plainToInstance } from 'class-transformer';
import { Region, SimpleAuthenticationDetailsProvider } from 'oci-common';
import { ObjectStorageClient } from 'oci-objectstorage';
import path from 'path';
import { BUCKETS } from '../constants';
import { UploadDocumentRequestDto } from '../dtos/request';
import { UploadResponseDto } from '../dtos/response';
import { generateNewFileName } from '../utils';

@Injectable()
export class OciService {
  private readonly ociClient: ObjectStorageClient;
  private readonly tenancyId: string = this.configService.getOrThrow('OCI_TENANCY_ID');
  private readonly userId: string = this.configService.getOrThrow('OCI_USER_ID');
  private readonly fingerprint: string = this.configService.getOrThrow('OCI_FINGERPRINT');
  private readonly privateKey: string = this.configService.getOrThrow('OCI_PRIVATE_KEY');
  private readonly passPhrase: string = this.configService.getOrThrow('OCI_PASSPHRASE');
  private readonly namespace: string = this.configService.getOrThrow('OCI_NAMESPACE');
  private readonly region: string = this.configService.getOrThrow('OCI_REGION');
  private readonly logger = new Logger(OciService.name);
  constructor(private configService: ConfigService) {
    this.ociClient = new ObjectStorageClient({
      authenticationDetailsProvider: new SimpleAuthenticationDetailsProvider(
        this.tenancyId,
        this.userId,
        this.fingerprint,
        Buffer.from(this.privateKey, 'base64').toString('utf-8'),
        this.passPhrase,
        Region.fromRegionId(this.region),
      ),
    });
  }

  async uploadFile(file: Express.Multer.File, { documentType }: UploadDocumentRequestDto): Promise<UploadResponseDto> {
    const bucketName = BUCKETS[documentType];
    const objectName = generateNewFileName(file.originalname);

    if (!bucketName) {
      this.logger.error('Could not find bucket name for document type', documentType);
      throw new BadRequestException('DOCUMENT.TYPE_NOT_SUPPORTED');
    }

    this.logger.debug(`Uploading file to bucket ${bucketName} with object name ${objectName}`);
    await this.ociClient.putObject({
      namespaceName: this.namespace,
      bucketName,
      putObjectBody: file.buffer,
      contentLength: file.buffer.length,
      objectName,
    });

    this.logger.log(`File uploaded successfully to bucket ${bucketName} with object name ${objectName}`);

    return plainToInstance(UploadResponseDto, {
      name: objectName,
      extension: path.extname(file.originalname),
      url: `https://objectstorage.${this.region}.oraclecloud.com/n/${this.namespace}/b/${bucketName}/o/${objectName}`,
      documentType,
    });
  }
}
